/* SPDX-License-Identifier: LGPL-2.1 */
/* Copyright (C) 2020-2021, Intel Corporation. All rights reserved. */
#ifndef _LIBCXL_H_
#define _LIBCXL_H_

#include <stdbool.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <ccan/short_types/short_types.h>

#ifdef HAVE_UUID
#include <uuid/uuid.h>
#else
typedef unsigned char uuid_t[16];
#endif

#ifdef __cplusplus
extern "C" {
#endif

#define FW_BYTE_ALIGN 128
#define FW_BLOCK_SIZE 128
typedef unsigned char fwblock[FW_BLOCK_SIZE];

struct cxl_ctx;
struct cxl_ctx *cxl_ref(struct cxl_ctx *ctx);
void cxl_unref(struct cxl_ctx *ctx);
int cxl_new(struct cxl_ctx **ctx);
void cxl_set_log_fn(struct cxl_ctx *ctx,
		void (*log_fn)(struct cxl_ctx *ctx, int priority,
			const char *file, int line, const char *fn,
			const char *format, va_list args));
int cxl_get_log_priority(struct cxl_ctx *ctx);
void cxl_set_log_priority(struct cxl_ctx *ctx, int priority);
void cxl_set_userdata(struct cxl_ctx *ctx, void *userdata);
void *cxl_get_userdata(struct cxl_ctx *ctx);
void cxl_set_private_data(struct cxl_ctx *ctx, void *data);
void *cxl_get_private_data(struct cxl_ctx *ctx);

struct cxl_memdev;
struct cxl_memdev *cxl_memdev_get_first(struct cxl_ctx *ctx);
struct cxl_memdev *cxl_memdev_get_next(struct cxl_memdev *memdev);
int cxl_memdev_get_id(struct cxl_memdev *memdev);
const char *cxl_memdev_get_devname(struct cxl_memdev *memdev);
int cxl_memdev_get_major(struct cxl_memdev *memdev);
int cxl_memdev_get_minor(struct cxl_memdev *memdev);
struct cxl_ctx *cxl_memdev_get_ctx(struct cxl_memdev *memdev);
unsigned long long cxl_memdev_get_pmem_size(struct cxl_memdev *memdev);
unsigned long long cxl_memdev_get_ram_size(struct cxl_memdev *memdev);
const char *cxl_memdev_get_firmware_verison(struct cxl_memdev *memdev);
size_t cxl_memdev_get_lsa_size(struct cxl_memdev *memdev);
int cxl_memdev_is_active(struct cxl_memdev *memdev);
int cxl_memdev_zero_lsa(struct cxl_memdev *memdev);
int cxl_memdev_get_lsa(struct cxl_memdev *memdev, void *buf, size_t length,
		size_t offset);
int cxl_memdev_set_lsa(struct cxl_memdev *memdev, void *buf, size_t length,
		size_t offset);
int cxl_memdev_cmd_identify(struct cxl_memdev *memdev);
int cxl_memdev_device_info_get(struct cxl_memdev *memdev);
int cxl_memdev_get_fw_info(struct cxl_memdev *memdev, bool is_os_img);
int cxl_memdev_transfer_fw(struct cxl_memdev *memdev, u8 action,
	u8 slot, u32 offset, int size, unsigned char *data, u32 transfer_fw_opcode);
int cxl_memdev_activate_fw(struct cxl_memdev *memdev, u8 action,
	u8 slot);
int cxl_memdev_get_supported_logs(struct cxl_memdev *memdev);
int cxl_memdev_get_log(struct cxl_memdev *memdev, const char *uuid, const unsigned int data_size);
int cxl_memdev_get_event_interrupt_policy(struct cxl_memdev *memdev);
int cxl_memdev_set_event_interrupt_policy(struct cxl_memdev *memdev, u32 int_policy);
int cxl_memdev_get_timestamp(struct cxl_memdev *memdev);
int cxl_memdev_set_timestamp(struct cxl_memdev *memdev, u64 timestamp);
int cxl_memdev_get_alert_config(struct cxl_memdev *memdev);
int cxl_memdev_set_alert_config(struct cxl_memdev *memdev, u32 alert_prog_threshold,
    u32 device_temp_threshold, u32 mem_error_threshold);
int cxl_memdev_get_health_info(struct cxl_memdev *memdev);
int cxl_memdev_get_event_records(struct cxl_memdev *memdev, u8 event_log_type);
int cxl_memdev_get_ld_info(struct cxl_memdev *memdev);
int cxl_memdev_ddr_info(struct cxl_memdev *memdev, u8 ddr_id);
int cxl_memdev_clear_event_records(struct cxl_memdev *memdev, u8 event_log_type,
    u8 clear_event_flags, u8 no_event_record_handles, u16 *event_record_handles);
int cxl_memdev_hct_start_stop_trigger(struct cxl_memdev *memdev,
	u8 hct_inst, u8 buf_control);
int cxl_memdev_hct_get_buffer_status(struct cxl_memdev *memdev,
	u8 hct_inst);
int cxl_memdev_hct_enable(struct cxl_memdev *memdev, u8 hct_inst);
int cxl_memdev_ltmon_capture_clear(struct cxl_memdev *memdev, u8 cxl_mem_id);
int cxl_memdev_ltmon_capture(struct cxl_memdev *memdev, u8 cxl_mem_id,
	u8 capt_mode, u16 ignore_sub_chg, u8 ignore_rxl0_chg, u8 trig_src_sel);
int cxl_memdev_ltmon_capture_freeze_and_restore(struct cxl_memdev *memdev,
	u8 cxl_mem_id, u8 freeze_restore);
int cxl_memdev_ltmon_l2r_count_dump(struct cxl_memdev *memdev,
	u8 cxl_mem_id);
int cxl_memdev_ltmon_l2r_count_clear(struct cxl_memdev *memdev,
	u8 cxl_mem_id);
int cxl_memdev_ltmon_basic_cfg(struct cxl_memdev *memdev, u8 cxl_mem_id,
	u8 tick_cnt, u8 global_ts);
int cxl_memdev_ltmon_watch(struct cxl_memdev *memdev, u8 cxl_mem_id,
	u8 watch_id, u8 watch_mode, u8 src_maj_st, u8 src_min_st, u8 src_l0_st,
	u8 dst_maj_st, u8 dst_min_st, u8 dst_l0_st);
int cxl_memdev_ltmon_capture_stat(struct cxl_memdev *memdev, u8 cxl_mem_id);
int cxl_memdev_ltmon_capture_log_dmp(struct cxl_memdev *memdev,
	u8 cxl_mem_id, u16 dump_idx, u16 dump_cnt);
int cxl_memdev_ltmon_capture_trigger(struct cxl_memdev *memdev,
	u8 cxl_mem_id, u8 trig_src);
int cxl_memdev_ltmon_enable(struct cxl_memdev *memdev, u8 cxl_mem_id,
	u8 enable);
int cxl_memdev_osa_os_type_trig_cfg(struct cxl_memdev *memdev,
	u8 cxl_mem_id, u16 lane_mask, u8 lane_dir_mask, u8 rate_mask, u16 os_type_mask);
int cxl_memdev_osa_cap_ctrl(struct cxl_memdev *memdev, u8 cxl_mem_id,
	u16 lane_mask, u8 lane_dir_mask, u8 drop_single_os, u8 stop_mode,
	u8 snapshot_mode, u16 post_trig_num, u16 os_type_mask);
int cxl_memdev_osa_cfg_dump(struct cxl_memdev *memdev, u8 cxl_mem_id);
int cxl_memdev_osa_ana_op(struct cxl_memdev *memdev, u8 cxl_mem_id,
	u8 op);
int cxl_memdev_osa_status_query(struct cxl_memdev *memdev, u8 cxl_mem_id);
int cxl_memdev_osa_access_rel(struct cxl_memdev *memdev, u8 cxl_mem_id);
int cxl_memdev_perfcnt_mta_ltif_set(struct cxl_memdev *memdev,
	u32 counter, u32 match_value, u32 opcode, u32 meta_field, u32 meta_value);
int cxl_memdev_perfcnt_mta_get(struct cxl_memdev *memdev, u8 type,
	u32 counter);
int cxl_memdev_perfcnt_mta_latch_val_get(struct cxl_memdev *memdev,
	u8 type, u32 counter);
int cxl_memdev_perfcnt_mta_counter_clear(struct cxl_memdev *memdev,
	u8 type, u32 counter);
int cxl_memdev_perfcnt_mta_cnt_val_latch(struct cxl_memdev *memdev,
	u8 type, u32 counter);
int cxl_memdev_perfcnt_mta_hif_set(struct cxl_memdev *memdev, u32 counter,
	u32 match_value, u32 addr, u32 req_ty, u32 sc_ty);
int cxl_memdev_perfcnt_mta_hif_cfg_get(struct cxl_memdev *memdev,
	u32 counter);
int cxl_memdev_perfcnt_mta_hif_latch_val_get(struct cxl_memdev *memdev,
	u32 counter);
int cxl_memdev_perfcnt_mta_hif_counter_clear(struct cxl_memdev *memdev,
	u32 counter);
int cxl_memdev_perfcnt_mta_hif_cnt_val_latch(struct cxl_memdev *memdev,
	u32 counter);
int cxl_memdev_perfcnt_ddr_generic_select(struct cxl_memdev *memdev,
	u8 ddr_id, u8 cid, u8 rank, u8 bank, u8 bankgroup, u64 event);
int cxl_memdev_perfcnt_ddr_generic_capture(struct cxl_memdev *memdev,
	u8 ddr_id, u32 poll_period_ms);
int cxl_memdev_perfcnt_ddr_dfi_capture(struct cxl_memdev *memdev,
	u8 ddr_id, u32 poll_period_ms);
int cxl_memdev_err_inj_drs_poison(struct cxl_memdev *memdev, u8 ch_id,
	u8 duration, u8 inj_mode, u16 tag);
int cxl_memdev_err_inj_drs_ecc(struct cxl_memdev *memdev, u8 ch_id,
	u8 duration, u8 inj_mode, u16 tag);
int cxl_memdev_err_inj_rxflit_crc(struct cxl_memdev *memdev, u8 cxl_mem_id);
int cxl_memdev_err_inj_txflit_crc(struct cxl_memdev *memdev, u8 cxl_mem_id);
int cxl_memdev_err_inj_viral(struct cxl_memdev *memdev, u8 ld_id);
int cxl_memdev_eh_eye_cap_run(struct cxl_memdev *memdev, u8 depth,
	u32 lane_mask);
int cxl_memdev_eh_eye_cap_read(struct cxl_memdev *memdev, u8 lane_id,
	u8 bin_num);
int cxl_memdev_eh_eye_cap_timeout_enable(struct cxl_memdev *memdev, u8 enable);
int cxl_memdev_eh_eye_cap_status(struct cxl_memdev *memdev);
int cxl_memdev_eh_adapt_get(struct cxl_memdev *memdev, u32 lane_id);
int cxl_memdev_eh_adapt_oneoff(struct cxl_memdev *memdev, u32 lane_id,
	u32 preload, u32 loops, u32 objects);
int cxl_memdev_eh_adapt_force(struct cxl_memdev *memdev, u32 lane_id,
	u32 rate, u32 vdd_bias, u32 ssc, u8 pga_gain, u8 pga_a0, u8 pga_off,
	u8 cdfe_a2, u8 cdfe_a3, u8 cdfe_a4, u8 cdfe_a5, u8 cdfe_a6, u8 cdfe_a7,
	u8 cdfe_a8, u8 cdfe_a9, u8 cdfe_a10, u16 dc_offset, u16 zobel_dc_offset,
	u16 udfe_thr_0, u16 udfe_thr_1, u16 median_amp, u8 zobel_a_gain,
	u8 ph_ofs_t);
int cxl_memdev_hbo_status(struct cxl_memdev *memdev, u8 print_output);
int cxl_memdev_hbo_transfer_fw(struct cxl_memdev *memdev);
int cxl_memdev_hbo_activate_fw(struct cxl_memdev *memdev);
int cxl_memdev_health_counters_clear(struct cxl_memdev *memdev,
	u32 bitmask);
int cxl_memdev_health_counters_get(struct cxl_memdev *memdev);
int cxl_memdev_hct_get_plat_param(struct cxl_memdev *memdev);
int cxl_memdev_err_inj_hif_poison(struct cxl_memdev *memdev, u8 ch_id,
	u8 duration, u8 inj_mode, u64 address);
int cxl_memdev_err_inj_hif_ecc(struct cxl_memdev *memdev, u8 ch_id,
	u8 duration, u8 inj_mode, u64 address);
int cxl_memdev_eh_link_dbg_cfg(struct cxl_memdev *memdev, u8 port_id, u8 op_mode,
	u8 cap_type, u16 lane_mask, u8 rate_mask, u32 timer_us, u32 cap_delay_us, u8 max_cap);
int cxl_memdev_eh_link_dbg_entry_dump(struct cxl_memdev *memdev, u8 entry_idx);
int cxl_memdev_eh_link_dbg_lane_dump(struct cxl_memdev *memdev, u8 entry_idx, u8 lane_idx);
int cxl_memdev_eh_link_dbg_reset(struct cxl_memdev *memdev);
int cxl_memdev_fbist_stopconfig_set(struct cxl_memdev *memdev,
	u32 fbist_id, u8 stop_on_wresp, u8 stop_on_rresp, u8 stop_on_rdataerr);
int cxl_memdev_fbist_cyclecount_set(struct cxl_memdev *memdev,
	u32 fbist_id, u8 txg_nr, u64 cyclecount);
int cxl_memdev_fbist_reset_set(struct cxl_memdev *memdev, u32 fbist_id,
	u8 txg0_reset, u8 txg1_reset);
int cxl_memdev_fbist_run_set(struct cxl_memdev *memdev, u32 fbist_id,
	u8 txg0_run, u8 txg1_run);
int cxl_memdev_fbist_run_get(struct cxl_memdev *memdev, u32 fbist_id);
int cxl_memdev_fbist_xfer_rem_cnt_get(struct cxl_memdev *memdev,
	u32 fbist_id, u8 thread_nr);
int cxl_memdev_fbist_last_exp_read_data_get(struct cxl_memdev *memdev,
	u32 fbist_id);
int cxl_memdev_fbist_curr_cycle_cnt_get(struct cxl_memdev *memdev,
	u32 fbist_id, u8 txg_nr);
int cxl_memdev_fbist_thread_status_get(struct cxl_memdev *memdev,
	u32 fbist_id, u8 txg_nr, u8 thread_nr);
int cxl_memdev_fbist_thread_trans_cnt_get(struct cxl_memdev *memdev,
	u32 fbist_id, u8 txg_nr, u8 thread_nr);
int cxl_memdev_fbist_thread_bandwidth_get(struct cxl_memdev *memdev,
	u32 fbist_id, u8 txg_nr, u8 thread_nr);
int cxl_memdev_fbist_thread_latency_get(struct cxl_memdev *memdev,
	u32 fbist_id, u8 txg_nr, u8 thread_nr);
int cxl_memdev_fbist_thread_perf_mon_set(struct cxl_memdev *memdev,
	u32 fbist_id, u8 txg_nr, u8 thread_nr, u8 pmon_preset_en, u8 pmon_clear_en,
	u8 pmon_rollover, u8 pmon_thread_lclk);
int cxl_memdev_fbist_top_read_status0_get(struct cxl_memdev *memdev,
	u32 fbist_id);
int cxl_memdev_fbist_top_err_cnt_get(struct cxl_memdev *memdev,
	u32 fbist_id);
int cxl_memdev_fbist_last_read_addr_get(struct cxl_memdev *memdev,
	u32 fbist_id);
int cxl_memdev_fbist_test_simpledata(struct cxl_memdev *memdev,
	u32 fbist_id, u8 test_nr, u64 start_address, u64 num_bytes);
int cxl_memdev_fbist_test_addresstest(struct cxl_memdev *memdev,
	u32 fbist_id, u8 test_nr, u64 start_address, u64 num_bytes, u32 seed);
int cxl_memdev_fbist_test_movinginversion(struct cxl_memdev *memdev,
	u32 fbist_id, u8 test_nr, u8 phase_nr, u64 start_address, u64 num_bytes,
	u32 ddrpage_size);
int cxl_memdev_fbist_test_randomsequence(struct cxl_memdev *memdev,
	u32 fbist_id, u8 phase_nr, u64 start_address, u64 num_bytes, u32 ddrpage_size,
	u32 seed_dr0, u32 seed_dr1);
int cxl_memdev_conf_read(struct cxl_memdev *memdev, u32 offset,
	u32 length);
int cxl_memdev_hct_get_config(struct cxl_memdev *memdev, u8 hct_inst);
int cxl_memdev_hct_read_buffer(struct cxl_memdev *memdev, u8 hct_inst, u8 num_entries_to_read);
int cxl_memdev_hct_set_config(struct cxl_memdev *memdev, u8 hct_inst, u8 config_flags,
	u8 port_trig_depth, u8 ignore_invalid, int filesize, u8 *trig_config_buffer);
int cxl_memdev_osa_os_patt_trig_cfg(struct cxl_memdev *memdev,
	u8 cxl_mem_id, u16 lane_mask, u8 lane_dir_mask, u8 rate_mask, u32 *patt_val,
	u32 *patt_mask);
int cxl_memdev_osa_misc_trig_cfg(struct cxl_memdev *memdev, u8 cxl_mem_id,
	u8 trig_en_mask);
int cxl_memdev_osa_data_read(struct cxl_memdev *memdev, u8 cxl_mem_id,
	u8 lane_id, u8 lane_dir, u16 start_entry, u8 num_entries);
int cxl_memdev_dimm_spd_read(struct cxl_memdev *memdev, u32 spd_id,
	u32 offset, u32 num_bytes);
int cxl_memdev_ddr_training_status(struct cxl_memdev *memdev);
int cxl_memdev_dimm_slot_info(struct cxl_memdev *memdev);
int cxl_memdev_pmic_vtmon_info(struct cxl_memdev *memdev);
int cxl_memdev_ddr_margin_run(struct cxl_memdev *memdev, u8 slice_num, u8 rd_wr_margin, u8 ddr_id);
int cxl_memdev_ddr_margin_status(struct cxl_memdev *memdev);
int cxl_memdev_ddr_margin_get(struct cxl_memdev *memdev);
int cxl_memdev_ddr_stats_run(struct cxl_memdev *memdev, u8 ddr_id,
							u32 monitor_time, u32 loop_count);
int cxl_memdev_ddr_stats_status(struct cxl_memdev *memdev, int* run_status, uint32_t* loop_count);
int cxl_memdev_ddr_stats_get(struct cxl_memdev *memdev);
int cxl_memdev_reboot_mode_set(struct cxl_memdev *memdev, u8 reboot_mode);
int cxl_memdev_curr_cxl_boot_mode_get(struct cxl_memdev *memdev);
int cxl_memdev_pcie_eye_run(struct cxl_memdev *memdev, u8 lane, u8 sw_scan, u8 ber);
int cxl_memdev_pcie_eye_status(struct cxl_memdev *memdev);
int cxl_memdev_pcie_eye_get_sw(struct cxl_memdev *memdev, uint offset);
int cxl_memdev_pcie_eye_get_hw(struct cxl_memdev *memdev);
int cxl_memdev_pcie_eye_get_sw_ber(struct cxl_memdev *memdev);
int cxl_memdev_get_cxl_link_status(struct cxl_memdev *memdev);
int cxl_memdev_get_device_info(struct cxl_memdev *memdev);
int cxl_memdev_read_ddr_temp(struct cxl_memdev *memdev);
int cxl_memdev_cxl_hpa_to_dpa(struct cxl_memdev *memdev, u64 hpa_address);
int cxl_memdev_get_cxl_membridge_errors(struct cxl_memdev *memdev);
int cxl_memdev_get_ddr_bw(struct cxl_memdev *memdev, u32 timeout, u32 iterations);
int cxl_memdev_get_ddr_latency(struct cxl_memdev *memdev, u32 measure_time);
int cxl_memdev_i2c_read(struct cxl_memdev *memdev, u16 slave_addr, u8 reg_addr, u8 num_bytes);
int cxl_memdev_i2c_write(struct cxl_memdev *memdev, u16 slave_addr, u8 reg_addr, u8 data);
int cxl_memdev_get_ddr_ecc_err_info(struct cxl_memdev *memdev);
int cxl_memdev_start_ddr_ecc_scrub(struct cxl_memdev *memdev);
int cxl_memdev_ddr_ecc_scrub_status(struct cxl_memdev *memdev);
int cxl_memdev_ddr_cont_scrub_status(struct cxl_memdev *memdev);
int cxl_memdev_ddr_cont_scrub_set(struct cxl_memdev *memdev, u32 cont_scrub_status);
int cxl_memdev_ddr_init_status(struct cxl_memdev *memdev);
int cxl_memdev_get_cxl_membridge_stats(struct cxl_memdev *memdev);
int cxl_memdev_trigger_coredump(struct cxl_memdev *memdev);
int cxl_memdev_ddr_err_inj_en(struct cxl_memdev *memdev, u32 ddr_id, u32 err_type, u64 ecc_fwc_mask);
int cxl_memdev_ddr_dimm_level_training_status(struct cxl_memdev *memdev);
int cxl_memdev_ddr_param_set(struct cxl_memdev *memdev, u32 ddr_interleave_sz,
	u32 ddr_interleave_ctrl_choice);
int cxl_memdev_ddr_param_get(struct cxl_memdev *memdev);
int cxl_memdev_core_volt_set(struct cxl_memdev *memdev, float core_volt);
int cxl_memdev_core_volt_get(struct cxl_memdev *memdev);
int cxl_memdev_oem_err_inj_viral(struct cxl_memdev *memdev, u32 viral_type);
int cxl_memdev_err_inj_ll_poison(struct cxl_memdev *memdev, u32 en_dis, u32 ll_err_type);
int cxl_memdev_pci_err_inj(struct cxl_memdev *memdev, u32 en_dis, u32 type, u32 err, u32 count, u32 opt1, u32 opt2);
int cxl_memdev_read_ltssm_states(struct cxl_memdev *memdev);
int cxl_memdev_ddr_page_select_set(struct cxl_memdev *memdev, u32 page_select_option);
int cxl_memdev_ddr_page_select_get(struct cxl_memdev *memdev);

#define cxl_memdev_foreach(ctx, memdev) \
        for (memdev = cxl_memdev_get_first(ctx); \
             memdev != NULL; \
             memdev = cxl_memdev_get_next(memdev))

struct cxl_cmd;
const char *cxl_cmd_get_devname(struct cxl_cmd *cmd);
struct cxl_cmd *cxl_cmd_new_raw(struct cxl_memdev *memdev, int opcode);
int cxl_cmd_set_input_payload(struct cxl_cmd *cmd, void *in, int size);
int cxl_cmd_set_output_payload(struct cxl_cmd *cmd, void *out, int size);
void cxl_cmd_ref(struct cxl_cmd *cmd);
void cxl_cmd_unref(struct cxl_cmd *cmd);
int cxl_cmd_submit(struct cxl_cmd *cmd);
int cxl_cmd_get_mbox_status(struct cxl_cmd *cmd);
int cxl_cmd_get_out_size(struct cxl_cmd *cmd);
struct cxl_cmd *cxl_cmd_new_identify(struct cxl_memdev *memdev);
int cxl_cmd_identify_get_fw_rev(struct cxl_cmd *cmd, char *fw_rev, int fw_len);
unsigned long long cxl_cmd_identify_get_partition_align(struct cxl_cmd *cmd);
unsigned int cxl_cmd_identify_get_lsa_size(struct cxl_cmd *cmd);
struct cxl_cmd *cxl_cmd_new_get_health_info(struct cxl_memdev *memdev);
int cxl_cmd_get_health_info_get_health_status(struct cxl_cmd *cmd);
int cxl_cmd_get_health_info_get_media_status(struct cxl_cmd *cmd);
int cxl_cmd_get_health_info_get_ext_status(struct cxl_cmd *cmd);
int cxl_cmd_get_health_info_get_life_used(struct cxl_cmd *cmd);
int cxl_cmd_get_health_info_get_temperature(struct cxl_cmd *cmd);
int cxl_cmd_get_health_info_get_dirty_shutdowns(struct cxl_cmd *cmd);
int cxl_cmd_get_health_info_get_volatile_errors(struct cxl_cmd *cmd);
int cxl_cmd_get_health_info_get_pmem_errors(struct cxl_cmd *cmd);
struct cxl_cmd *cxl_cmd_new_get_lsa(struct cxl_memdev *memdev,
		unsigned int offset, unsigned int length);
void *cxl_cmd_get_lsa_get_payload(struct cxl_cmd *cmd);
struct cxl_cmd *cxl_cmd_new_set_lsa(struct cxl_memdev *memdev,
		void *buf, unsigned int offset, unsigned int length);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif
